import codecs
import multiprocessing as mp
from multiprocessing.context import SpawnProcess
from multiprocessing.synchronize import Lock
import os
from pathlib import Path
from queue import Empty
import sys
from typing import Dict, Generator, List, NamedTuple, Optional, Set, Tuple

import polars as pl

import pod5 as p5
from pod5.tools.parsers import prepare_pod5_view_argparser, run_tool
from pod5.tools.polars_utils import pl_format_empty_string, pl_format_read_id
from pod5.tools.utils import (
    DEFAULT_THREADS,
    collect_inputs,
    init_logging,
    limit_threads,
    logged,
    logged_all,
    terminate_processes,
)


logger = init_logging()

pl.enable_string_cache(True)


class Field(NamedTuple):
    """Container class for storing the polars expression for a named field"""

    expr: pl.Expr
    docs: str


# This dict defines the order of the fields
FIELDS: Dict[str, Field] = {
    "read_id": Field(pl.col("read_id"), "Read UUID"),
    "filename": Field(pl.col("filename"), "Source pod5 filename"),
    "read_number": Field(pl.col("read_number"), "Read number"),
    "channel": Field(pl.col("channel"), "1-indexed channel"),
    "mux": Field(pl.col("mux"), "1-indexed well"),
    "end_reason": Field(pl.col("end_reason"), "End reason string"),
    "start_time": Field(
        pl.col("start_time"),
        "Seconds since the run start to the first sample of this read",
    ),
    "start_sample": Field(
        pl.col("start").alias("start_sample"),
        "Samples recorded on this channel since run start to the first sample of this read",
    ),
    "duration": Field(pl.col("duration"), "Seconds of sampling for this read"),
    "num_samples": Field(pl.col("num_samples"), "Number of signal samples"),
    "minknow_events": Field(
        pl.col("minknow_events"),
        "Number of minknow events that this read contains",
    ),
    "sample_rate": Field(
        pl.col("sample_rate"), "Number of samples recorded each second"
    ),
    "median_before": Field(
        pl.col("median_before"), "Current level in this well before the read"
    ),
    "predicted_scaling_scale": Field(
        pl.col("predicted_scaling_scale"), "Scale for predicted read scaling"
    ),
    "predicted_scaling_shift": Field(
        pl.col("predicted_scaling_shift"), "Shift for predicted read scaling"
    ),
    "tracked_scaling_scale": Field(
        pl.col("tracked_scaling_scale"), "Scale for tracked read scaling"
    ),
    "tracked_scaling_shift": Field(
        pl.col("tracked_scaling_shift"), "Shift for tracked read scaling"
    ),
    "num_reads_since_mux_change": Field(
        pl.col("num_reads_since_mux_change"),
        "Number of selected reads since the last mux change on this channel",
    ),
    "time_since_mux_change": Field(
        pl.col("time_since_mux_change"),
        "Seconds since the last mux change on this channel",
    ),
    "run_id": Field(pl.col("protocol_run_id").alias("run_id"), "Run UUID"),
    "sample_id": Field(pl.col("sample_id"), "User-supplied name for the sample"),
    "experiment_id": Field(
        pl.col("experiment_id"), "User-supplied name for the experiment"
    ),
    "flow_cell_id": Field(pl.col("flow_cell_id"), "The flow cell id"),
    "pore_type": Field(pl.col("pore_type"), "Name of the pore in this well"),
}


@logged()
def print_fields():
    """Print a list of the available columns"""
    for name, field in FIELDS.items():
        print(f"{name.ljust(28)} {field.docs}")
    print("")


@logged_all
def get_field_or_raise(key: str) -> Field:
    """Get the Field for this key or raise a KeyError"""
    try:
        return FIELDS[key]
    except KeyError:
        raise KeyError(
            f"Field: '{key}' did not match any known fields. "
            "Please check it exists by viewing `-L/--list-fields`"
        )


@logged_all
def select_fields(
    *,
    group_read_id: bool = False,
    include: Optional[str] = None,
    exclude: Optional[str] = None,
) -> Set[str]:
    """Select fields to write"""
    selected: Set[str] = set([])

    # Select only read ids
    if group_read_id:
        selected.add("read_id")
        return selected

    if include:
        for key in include.split(","):
            key = key.strip()
            if not key:
                continue
            get_field_or_raise(key)
            selected.add(key)

    # Default selection - All fields
    if not selected:
        selected.update(FIELDS)

    if exclude:
        for key in exclude.split(","):
            key = key.strip()
            if not key:
                continue
            get_field_or_raise(key)
            try:
                selected.remove(key)
            except KeyError:
                pass

    if not selected:
        raise RuntimeError("Zero Fields selected. Please select at least one field")

    return selected


def format_view_table(
    lazyframe: pl.LazyFrame, path: Path, selected_fields: Set[str]
) -> pl.LazyFrame:
    """Format the view table based on the selected fields"""
    maybe_empty = ["experiment_id", "protocol_run_id", "sample_id", "flow_cell_id"]

    lazyframe = lazyframe.with_columns(
        # Add the source filename
        pl.lit(path.name).alias("filename"),
        pl_format_read_id(pl.col("read_id")),
        # Rename fields to better match legacy sequencing summary
        (pl.col("well").alias("mux")),
        (pl.col("num_minknow_events").alias("minknow_events")),
        pl.col("experiment_name").alias("experiment_id"),
        # Compute the start_time in seconds
        (pl.col("start") / pl.col("sample_rate")).alias("start_time"),
        # Compute the duration of the read in seconds
        (pl.col("num_samples") / pl.col("sample_rate")).alias("duration"),
    )

    # Replace potentially empty fields with "not_set"
    # This can't be done in the above expression due to the behaviour of
    # keep_name()
    lazyframe = lazyframe.with_columns(
        pl_format_empty_string(pl.col(maybe_empty), "not_set").keep_name()
    )

    # Apply the field selection
    lazyframe = lazyframe.select(
        field.expr for key, field in FIELDS.items() if key in selected_fields
    )

    return lazyframe


@logged(log_time=True)
def write(
    ldf: pl.LazyFrame,
    output: Optional[Path],
    separator: str = "\t",
) -> None:
    """Write the polars.LazyFrame"""

    kwargs = dict(
        has_header=False, separator=separator, null_value="", float_precision=8
    )

    # Write to the nominated output path
    if output is not None:
        with output.open("ab") as f:
            ldf.collect().write_csv(f, **kwargs)
        return

    # No output path, collect the table content as a string and print it to stdout
    content = ldf.collect().write_csv(**kwargs)
    try:
        # Do not add additional newline at the end, this ensures consistency with
        # writing to file
        print(content, end="")
    except BrokenPipeError as exc:
        # https://docs.python.org/3/library/signal.html#note-on-sigpipe
        devnull = os.open(os.devnull, os.O_WRONLY)
        os.dup2(devnull, sys.stdout.fileno())
        raise exc


def write_header(
    output: Optional[Path], selected: Set[str], separator: str = "\t"
) -> None:
    """Write the header line"""
    header = separator.join(key for key in FIELDS if key in selected)
    if output is None:
        print(header, file=sys.stdout)
    else:
        output.write_text(header + "\n")


@logged_all
def resolve_output(output: Optional[Path], force_overwrite: bool) -> Optional[Path]:
    """
    Resolve the output path if necessary checking for no accidental overwrite
    and resolving to default output if given a path
    """
    if output is None:
        return None

    # Do not allow accidental overwrite
    if output.is_file():
        if not force_overwrite:
            raise FileExistsError(
                f"{output} points to an existing file and --force-overwrite not set"
            )
        output.unlink()

    # If given a directory, check the default filename is valid
    if output.is_dir():
        default_name = output / "view.txt"
        return resolve_output(default_name, force_overwrite)

    return output


@logged()
def assert_unique_acquisition_id(run_info: pl.LazyFrame, path: Path) -> None:
    """
    Perform a check that the acquisition ids are unique raising AssertionError otherwise
    """
    groups = run_info.collect().group_by(pl.col("acquisition_id"))
    common_acq_ids = [acq_id for acq_id, frame in groups if frame.n_unique() != 1]
    if common_acq_ids:
        raise AssertionError(
            f"Found non-unique run_info acquisition_id in {path.name}: {common_acq_ids}. "
        )


def parse_reads_table_all(reader: p5.Reader) -> pl.LazyFrame:
    """
    Parse all records in the reads table returning a polars LazyFrame
    """
    logger.debug(f"Parsing {reader.path.name} records")
    read_table = reader.read_table.read_all()
    reads = (
        pl.from_arrow(read_table, rechunk=False)
        .drop(["signal"])
        .lazy()
        .with_columns(pl.col("run_info").cast(pl.Utf8))
    )
    return reads


def parse_reads_table_batch(
    reader: p5.Reader, batch_index: int
) -> Tuple[pl.LazyFrame, int]:
    """
    Parse the reads table record batch at `batch_index` from a pod5 file returning a
    polars LazyFrame and the number of records in it
    """
    logger.debug(f"Parsing {reader.path.name} record batch {batch_index}")
    read_table = reader.read_table.get_batch(batch_index)
    reads = (
        pl.from_arrow(read_table, rechunk=False)
        .drop(["signal"])
        .lazy()
        .with_columns(pl.col("run_info").cast(pl.Utf8))
    )
    return reads, read_table.num_rows


@logged_all
def parse_read_table_chunks(
    reader: p5.Reader, approx_size: int = 99_999
) -> Generator[pl.LazyFrame, None, None]:
    """
    Read record batches and yield polars lazyframes of `approx_size` records.
    Records are yielded in units of whole batches of the underlying table
    """
    chunks: List[pl.LazyFrame] = []
    chunk_rows = 0

    for batch_index in range(reader.read_table.num_record_batches):
        reads, n_rows = parse_reads_table_batch(reader, batch_index)

        chunks.append(reads)
        chunk_rows += n_rows

        if chunk_rows > approx_size:
            reads_chunk = pl.concat(chunks)
            logger.debug(f"Emitting chunk of {chunk_rows} rows")
            chunks = []
            chunk_rows = 0
            yield reads_chunk

    if chunk_rows > 0:
        reads_chunk = pl.concat(chunks)
        chunks = []
        logger.debug(f"Emitting final chunk of {chunk_rows} rows")
        yield reads_chunk


@logged()
def parse_run_info_table(reader: p5.Reader) -> pl.LazyFrame:
    """Parse the reads table from a pod5 file returning a polars LazyFrame"""
    run_info_table = reader.run_info_table.read_all().drop(
        ["context_tags", "tracking_id"]
    )
    run_info = pl.from_arrow(run_info_table, rechunk=False).lazy().unique()
    return run_info


@logged()
def join_reads_to_run_info(reads: pl.LazyFrame, run_info: pl.LazyFrame) -> pl.LazyFrame:
    """Join the reads and run_info tables"""
    return reads.join(
        run_info.unique(),
        left_on="run_info",
        right_on="acquisition_id",
    )


def get_reads_tables(
    path: Path, selected_fields: Set[str], threshold: int = 100_000
) -> Generator[pl.LazyFrame, None, None]:
    """
    Generate lazy dataframes from pod5 records. If the number of records
    is greater than `threshold` then yield chunks to limit memory consumption and
    improve overall performance
    """
    with p5.Reader(path) as reader:
        run_info = parse_run_info_table(reader)
        assert_unique_acquisition_id(run_info, path)

        if reader.num_reads <= threshold:
            reads_table = parse_reads_table_all(reader)
            joined = join_reads_to_run_info(reads_table, run_info)
            yield format_view_table(joined, path, selected_fields)
            return

        for reads_chunk in parse_read_table_chunks(reader, approx_size=threshold - 1):
            joined = join_reads_to_run_info(reads_chunk, run_info)
            yield format_view_table(joined, path, selected_fields)


def join_workers(processes: List[SpawnProcess], exceptions: mp.JoinableQueue) -> None:
    """Poll workers checking for exceptions which will likely cause"""
    prcs = {p for p in processes}
    while prcs:
        try:
            exc, path = exceptions.get(timeout=0.1)
            terminate_processes(processes)
            exceptions.task_done()
            if isinstance(exc, BrokenPipeError):
                sys.exit(1)
            else:
                terminate_processes(processes)
                raise RuntimeError(f"Error while processing '{path}'") from exc
        except Empty:
            pass

        done = set()
        for prc in prcs:
            exit_code = prc.exitcode

            if exit_code is None:
                continue

            if exit_code > 0:
                terminate_processes(processes)
                raise mp.ProcessError(
                    f"Unexpected exception ocurrecd in {prc} - exit code: {exit_code}"
                )
            else:
                done.add(prc)
        prcs.difference_update(done)

    for prc in processes:
        prc.join()


@logged_all
def worker_process(
    paths: mp.JoinableQueue,
    exceptions: mp.JoinableQueue,
    lock: Lock,
    output: Path,
    separator: bool,
    selection: Set[str],
) -> None:
    """
    Consume pod5 paths from `paths` queue, parse the records and write to `output` after
    acquiring `lock`.
    Returns `None` when all finish sentinel `None` is received in `paths` queue.
    """
    path: Optional[Path] = None
    try:
        while True:
            path = paths.get()
            if path is None:
                paths.task_done()
                break

            try:
                for table in get_reads_tables(path, selection):
                    with lock:
                        write(ldf=table, output=output, separator=separator)
            finally:
                paths.task_done()
        paths.close()

    except Exception as exc:
        exceptions.put((exc, path))


def launch_view_workers(
    paths: Set[Path],
    output: Path,
    selection: Set[str],
    separator: str,
    num_workers: int,
):
    ctx = mp.get_context("spawn")
    write_lock = ctx.Lock()
    paths_queue = ctx.JoinableQueue(maxsize=len(paths) * 2)
    exceptions_queue = ctx.JoinableQueue(maxsize=len(paths))

    # Prepare the paths queue
    for path in paths:
        paths_queue.put(path)

    processes: List[SpawnProcess] = []
    for _ in range(num_workers):
        worker = ctx.Process(
            target=worker_process,
            kwargs=dict(
                paths=paths_queue,
                exceptions=exceptions_queue,
                lock=write_lock,
                output=output,
                separator=separator,
                selection=selection,
            ),
            daemon=True,
        )
        worker.start()
        processes.append(worker)

        # Enqueue a stop sentinel for each worker
        paths_queue.put(None)

    join_workers(processes, exceptions_queue)

    paths_queue.join()
    paths_queue.close()
    paths_queue.join_thread()


@logged_all
def view_pod5(
    inputs: List[Path],
    output: Path,
    separator: str = "\t",
    recursive: bool = False,
    force_overwrite: bool = False,
    list_fields: bool = False,
    no_header: bool = False,
    threads: int = DEFAULT_THREADS,
    **kwargs,
) -> None:
    """Given a list of POD5 files write a table to view their contents"""

    if list_fields:
        print_fields()
        return

    threads = limit_threads(threads)

    output_path = resolve_output(output, force_overwrite)

    # Decode escaped separator characters e.g. \t
    sep = codecs.decode(separator, "unicode-escape")

    # Parse column selection args
    selection = select_fields(**kwargs)

    collected_paths = collect_inputs(
        inputs, recursive=recursive, pattern="*.pod5", threads=threads
    )
    if not collected_paths:
        raise AssertionError("Found no pod5 files searching inputs")

    num_workers = min(len(collected_paths), threads)

    if not no_header:
        write_header(output=output_path, selected=selection, separator=sep)

    launch_view_workers(
        paths=collected_paths,
        output=output_path,
        selection=selection,
        separator=sep,
        num_workers=num_workers,
    )


def main():
    run_tool(prepare_pod5_view_argparser())


if __name__ == "__main__":
    main()



"""
pod5 setup.py
Proprietary and confidential information of Oxford Nanopore Technologies plc
All rights reserved; (c)2022: Oxford Nanopore Technologies plc

This script can either install a development version of pod5 to the current
Python environment, or create a Python wheel.

"""


import setuptools

if __name__ == "__main__":
    setuptools.setup()



"""
Utility functions for pod5 tools
"""

import datetime
import functools
import glob
import logging
import multiprocessing as mp
from multiprocessing.context import SpawnProcess
import os
from time import perf_counter
from typing import Collection, Iterable, List, Set, Union
from pathlib import Path
import uuid


# os.cpu_count() can return None if it fails
DEFAULT_THREADS = min(os.cpu_count() or 4, 4)


def init_logging():
    """Initialise logging only if POD5_DEBUG is true"""
    if not is_pod5_debug():
        logger = logging.getLogger("pod5")
        logger.addHandler(logging.NullHandler())
    else:
        datetime_now = datetime.datetime.now().strftime("%Y-%m-%d--%H-%M-%S")
        if mp.current_process().name == "MainProcess":
            pid = "main"
        else:
            pid = f"p-{os.getpid()}"
        logging.basicConfig(
            level=logging.DEBUG,
            filename=f"{datetime_now}-{pid}-pod5.log",
            format="%(asctime)s %(levelname)s %(message)s",
        )
        logger = logging.getLogger("pod5")
    return logger


def logged(log_return: bool = False, log_args: bool = False, log_time: bool = False):
    """Logging parameterised decorator"""
    log_fn = logging.debug

    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            uid = f"{str(uuid.uuid4())[:2]}:'{func.__name__}'"
            if log_args:
                log_fn("{0}:{1}, {2}".format(uid, args, kwargs))
            else:
                log_fn("{0}".format(uid))
            try:
                started = perf_counter()
                ret = func(*args, **kwargs)
            except Exception as exc:
                log_fn("{0}:Exception:{1}".format(uid, exc))
                raise exc
            if log_time:
                duration_s = perf_counter() - started
                log_fn("{0}:Done:{1:.3f}s".format(uid, duration_s))
            if log_return:
                log_fn("{0}:Returned:{1}".format(uid, ret))
            return ret

        return wrapper

    return decorator


logged_all = logged(log_return=True, log_args=True, log_time=True)


@logged_all
def terminate_processes(processes: List[SpawnProcess]) -> None:
    """terminate all child processes"""
    for proc in processes:
        try:
            proc.terminate()
        except ValueError:
            # Catch ValueError raised if proc is already closed
            pass
    return


@logged(log_return=True)
def limit_threads(requested: int) -> int:
    """
    Santise and limit the number of ``requested`` threads to the number of logical cores
    """
    if requested < 1:
        return os.cpu_count() or 4
    return min(os.cpu_count() or requested, requested)


@logged(log_time=True)
def collect_inputs(
    paths: Iterable[Path],
    recursive: bool,
    pattern: Union[str, Collection[str]],
    threads: int = DEFAULT_THREADS,
) -> Set[Path]:
    """
    Returns a set of `path` which match any of the given glob-style `pattern`s

    If a path is a directory this will be globbed (optionally recursively).
    If a path is a file then it must also match any of the given `pattern`s.

    Raises FileExistsError if any inputs do not exist
    """
    paths = set(paths)
    assert_inputs_exist(paths)
    if len(paths) == 0:
        raise AssertionError("Got 0 input paths to search")

    return search_paths(paths, recursive, pattern, min(threads, len(paths)))


@logged(log_time=True, log_return=True)
def assert_inputs_exist(inputs: Iterable[Path]):
    """Assert all inputs exist. Raises FileExistsError otherwise"""
    bad_paths = set()
    for path in set(inputs):
        if not path.exists():
            bad_paths.add(path)

    if bad_paths:
        raise FileExistsError(f"{len(bad_paths)} inputs do not exist: {bad_paths}")


@logged(log_time=True)
def search_paths(
    paths: Iterable[Path],
    recursive: bool,
    pattern: Union[str, Collection[str]],
    threads: int = DEFAULT_THREADS,
) -> Set[Path]:
    """
    Search all `paths` matching any of `patterns` searching directories recursively
    if requested
    """
    if isinstance(pattern, str):
        pattern = [pattern]

    srch = functools.partial(search_path, recursive=recursive, patterns=pattern)

    all_matches: Set[Path] = set()
    with mp.Pool(processes=threads) as pool:
        for matches in pool.imap_unordered(srch, paths):
            all_matches.update(matches)

    return all_matches


@logged(log_time=True)
def search_path(path: Path, recursive: bool, patterns: Collection[str]) -> Set[Path]:
    """
    Search `path` matching `pattern` searching directories recursively if requested
    """

    def _any_match(path: Path):
        return any(path.match(p) for p in patterns)

    # Get the recursive or non-recursive glob function.
    matching_files = set()
    if path.is_dir():
        pattern = str(path / "**" / "*") if recursive else str(path / "*")
        for matching_pathname in glob.glob(pattern, recursive=recursive):
            matching_path = Path(matching_pathname)
            if matching_path.is_file() and _any_match(matching_path):
                matching_files.add(matching_path)

    # Non-directory, assert that it is a file and that it matches the file_pattern
    elif path.is_file() and _any_match(path):
        matching_files.add(path)

    return matching_files


@logged(log_time=True)
def assert_no_duplicate_filenames(inputs: Collection[Path]) -> None:
    """
    Raises ValueError if there are duplicate filenames in the collection of Paths
    """
    names = [path.name for path in inputs]
    if len(names) != len(set(names)):
        raise ValueError(
            "One or more inputs share the same filename. "
            "This would cause a files to be overwritten at runtime"
        )


# Do not log this function as it's executed at import time
def is_disable_pbar() -> bool:
    """Check if POD5_PBAR is set returning true if PBAR should be disabled"""
    try:
        enabled = bool(int(os.environ.get("POD5_PBAR", "1")))
        return not enabled
    except Exception:
        return False


PBAR_DEFAULTS = dict(
    disable=is_disable_pbar(),
    smoothing=0.0,
    dynamic_ncols=True,
    ascii=True,
)


# Do not log this function as it's executed during logging initialisation
def is_pod5_debug() -> bool:
    """Check if POD5_DEBUG is set"""
    try:
        debug = bool(int(os.environ.get("POD5_DEBUG", "0")))
        return debug
    except Exception:
        return True



"""
Tool for recovering truncated pod5 files
"""
import dataclasses
import typing
from pathlib import Path

import lib_pod5 as p5b
import pod5 as p5
from pod5.tools.parsers import prepare_pod5_recover_argparser, run_tool
from pod5.tools.utils import collect_inputs


@dataclasses.dataclass
class RecoveredData:
    """
    Holds info about recovered data.
    """

    signal_rows: int = 0
    reads: int = 0
    run_infos: int = 0
    files_with_errors: int = 0


def do_consistency_check(path: Path, recovered: RecoveredData) -> bool:
    """
    Count the number of recoverd records in `path` updating `recovered`.
    Returns True is this file was successfully recovered
    """
    try:
        with p5.Reader(path) as file:
            for i in range(file.signal_table.num_record_batches):
                batch = file.signal_table.get_batch(i)
                recovered.signal_rows += batch.num_rows

            for i in range(file.run_info_table.num_record_batches):
                batch = file.run_info_table.get_batch(i)
                recovered.run_infos += batch.num_rows

            for i in range(file.read_table.num_record_batches):
                batch = file.read_table.get_batch(i)
                recovered.reads += batch.num_rows
    except RuntimeError:
        recovered.files_with_errors += 1
        return False
    return True


def is_file_ok(path: Path) -> bool:
    try:
        with p5.Reader(path):
            pass
        return True
    except RuntimeError:
        return False


def recover_pod5(inputs: typing.List[Path], force_overwrite: bool, recursive: bool):
    """
    Given a list of truncated pod5 files, recover their data.
    """

    paths = collect_inputs(inputs, recursive=recursive, pattern=["*.tmp", "*.pod5"])

    paths_to_recover = [p for p in paths if not is_file_ok(p)]

    if len(paths_to_recover) == 0:
        print(f"None of the {len(paths)} files given need recovery")
        return

    for path in paths_to_recover:
        dest = path.parent / (path.stem + "_recovered.pod5")
        if dest.exists():
            if force_overwrite:
                dest.unlink()
            else:
                raise FileExistsError(
                    f"Output files already exists and --force-overwrite not set. "
                    f"Refusing to overwrite {dest}."
                )

    recovered_data = RecoveredData()
    for input_file in paths_to_recover:
        dest = path.parent / (path.stem + "_recovered.pod5")
        # recover_file returns us an open writer:
        f = p5b.recover_file(str(input_file.resolve()), str(dest.resolve()))

        # We don't want to write any additional data:
        f.close()

        # Check how consistent the recovered file is:
        success = do_consistency_check(dest.resolve(), recovered_data)
        if not success:
            print(f"{dest} - Recovery failed")
            dest.unlink()
        else:
            print(f"{dest} - Recovered")

    print(
        f"Recovered {recovered_data.signal_rows} signal rows, "
        f"{recovered_data.reads} reads, "
        f"{recovered_data.run_infos} run infos, "
        f"{recovered_data.files_with_errors} files with errors"
    )


def main():
    run_tool(prepare_pod5_recover_argparser())


if __name__ == "__main__":
    main()



"""
Parsers for pod5 tools.

Each parser should have set_defaults(func=tool) called where tool is the core function
of a pod5 tool. These are commonly tool_pod5 (e.g. subset_pod5, merge_pod5)
"""

import argparse
import sys
from pathlib import Path
from typing import Any, Optional

from pod5.signal_tools import DEFAULT_SIGNAL_CHUNK_SIZE
from pod5.tools.utils import DEFAULT_THREADS, is_pod5_debug


class SubcommandHelpFormatter(
    argparse.RawDescriptionHelpFormatter,
    argparse.ArgumentDefaultsHelpFormatter,
):
    """
    Helper function to prettier print subcommand help. This removes some
    extra lines of output when a final command parser is not selected.
    """

    def _format_action(self, action):
        parts = super(SubcommandHelpFormatter, self)._format_action(action)
        if action.nargs == argparse.PARSER:
            parts = "\n".join(parts.split("\n")[1:])
        return parts


def run_tool(parser: argparse.ArgumentParser) -> Any:
    """Run the tool prepared by an argparser"""
    kwargs = vars(parser.parse_args())
    tool_func = kwargs.pop("func")
    try:
        return tool_func(**kwargs)
    except Exception as exc:
        if is_pod5_debug():
            raise exc
        print(f"\nPOD5 has encountered an error: '{exc}'", file=sys.stderr)
        print("\nFor detailed information set POD5_DEBUG=1'", file=sys.stderr)
        exit(1)


def add_recursive_argument(parser: argparse.ArgumentParser):
    parser.add_argument(
        "-r",
        "--recursive",
        default=False,
        action="store_true",
        help="Search for input files recursively matching `*.pod5`",
    )


def add_force_overwrite_argument(parser: argparse._ActionsContainer):
    parser.add_argument(
        "-f",
        "--force-overwrite",
        action="store_true",
        help="Overwrite destination files",
    )


#
# CONVERT - fast5
#
def pod5_convert_from_fast5_argparser(
    parent: Optional[argparse._SubParsersAction] = None,
) -> argparse.ArgumentParser:
    """
    Create an argument parser for the pod5 convert-from-fast5 tool
    """

    _desc = "Convert fast5 file(s) into a pod5 file(s)"

    if parent is None:
        parser = argparse.ArgumentParser(description=_desc)
    else:
        parser = parent.add_parser(
            name="fast5",
            aliases=["from_fast5"],
            description=_desc,
            formatter_class=SubcommandHelpFormatter,
        )

    parser.add_argument(
        "inputs", type=Path, nargs="+", help="Input path for fast5 file"
    )
    required_group = parser.add_argument_group("required arguments")
    required_group.add_argument(
        "-o",
        "--output",
        type=Path,
        required=True,
        help="Output path for the pod5 file(s). This can be an existing "
        "directory (creating 'output.pod5' within it) or a new named file path. "
        "A directory must be given when using --one-to-one.",
    )
    add_recursive_argument(parser)
    parser.add_argument(
        "-t",
        "--threads",
        default=DEFAULT_THREADS,
        type=int,
        help="Set the number of threads to use",
    )
    parser.add_argument(
        "--strict",
        action="store_true",
        help="Immediately quit if an exception is encountered during conversion "
        "instead of continuing with remaining inputs after issuing a warning",
    )
    output_group = parser.add_argument_group("output control arguments")
    output_group.add_argument(
        "-O",
        "--one-to-one",
        type=Path,
        default=None,
        help="Output files are written 1:1 to inputs. 1:1 output files are "
        "written to the output directory in a new directory structure relative to the "
        "directory path provided to this argument. This directory path must be a "
        "relative parent of all inputs.",
    )
    add_force_overwrite_argument(output_group)
    output_group.add_argument(
        "--signal-chunk-size",
        default=DEFAULT_SIGNAL_CHUNK_SIZE,
        help="Chunk size to use for signal data set",
        type=int,
    )

    def run(**kwargs):
        from pod5.tools.pod5_convert_from_fast5 import convert_from_fast5

        return convert_from_fast5(**kwargs)

    parser.set_defaults(func=run)

    return parser


#
# CONVERT - to_fast5
#
def pod5_convert_to_fast5_argparser(
    parent: Optional[argparse._SubParsersAction] = None,
) -> argparse.ArgumentParser:
    """
    Create an argument parser for the pod5 convert-to-fast5 tool
    """
    _desc = "Convert pod5 file(s) into fast5 file(s)"
    if parent is None:
        parser = argparse.ArgumentParser(description=_desc)
    else:
        parser = parent.add_parser(
            name="to_fast5",
            description=_desc,
            formatter_class=SubcommandHelpFormatter,
        )

    parser.add_argument("inputs", type=Path, nargs="+")
    required_group = parser.add_argument_group("required arguments")
    required_group.add_argument(
        "-o",
        "--output",
        type=Path,
        required=True,
        help="Output path for the pod5 file(s). This can be an existing "
        "directory (creating 'output.pod5' within it) or a new named file path. "
        "A directory must be given when using --one-to-one.",
    )
    add_recursive_argument(parser)
    parser.add_argument(
        "-t",
        "--threads",
        default=DEFAULT_THREADS,
        type=int,
        help="How many file writers to keep active",
    )
    output_group = parser.add_argument_group("output control arguments")
    add_force_overwrite_argument(output_group)
    output_group.add_argument(
        "--file-read-count",
        default=4000,
        type=int,
        help="Number of reads to write per file.",
    )

    def run(**kwargs):
        from pod5.tools.pod5_convert_to_fast5 import convert_to_fast5

        return convert_to_fast5(**kwargs)

    parser.set_defaults(func=run)

    return parser


#
# CONVERT - root
#
def prepare_pod5_convert(parent: argparse._SubParsersAction) -> argparse.ArgumentParser:
    """Create an argument paraser for the pod5 convert entry point"""

    _desc = "File conversion tools"

    convert_parser = parent.add_parser(
        name="convert",
        description=_desc,
        epilog="Example: pod5 convert fast5 input.fast5 --output output.pod5",
        formatter_class=SubcommandHelpFormatter,
    )
    convert_parser.set_defaults(func=lambda x: convert_parser.print_help())

    sub_convert_parser = convert_parser.add_subparsers(title="conversion type")

    pod5_convert_from_fast5_argparser(sub_convert_parser)
    pod5_convert_to_fast5_argparser(sub_convert_parser)

    return convert_parser


#
# Filter
#
def prepare_pod5_filter_argparser(
    parent: Optional[argparse._SubParsersAction] = None,
) -> argparse.ArgumentParser:
    """Create an argument parser for the pod5 filter tool"""

    _desc = "Take a subset of reads using a list of read_ids from one or more inputs"
    if parent is None:
        parser = argparse.ArgumentParser(description=_desc)
    else:
        parser = parent.add_parser(
            name="filter",
            description=_desc,
            epilog="Example: pod5 filter inputs*.pod5 --ids read_ids.txt --output filtered.pod5",
        )

    # Core arguments
    parser.add_argument(
        "inputs", type=Path, nargs="+", help="Pod5 filepaths to use as inputs"
    )
    add_recursive_argument(parser)
    add_force_overwrite_argument(parser)

    required_group = parser.add_argument_group("required arguments")
    required_group.add_argument(
        "-i",
        "--ids",
        type=Path,
        required=True,
        help="A file containing a list of only valid read ids to filter from inputs",
    )
    required_group.add_argument(
        "-o",
        "--output",
        type=Path,
        required=True,
        help="Destination output filename",
    )
    parser.add_argument(
        "-t",
        "--threads",
        type=int,
        default=DEFAULT_THREADS,
        help="Number of workers",
    )

    content_group = parser.add_argument_group("content settings")
    content_group.add_argument(
        "-M",
        "--missing-ok",
        action="store_true",
        help="Allow missing read_ids",
    )
    content_group.add_argument(
        "-D",
        "--duplicate-ok",
        action="store_true",
        help="Allow duplicate read_ids",
    )

    def run(**kwargs):
        from pod5.tools.pod5_filter import filter_pod5

        return filter_pod5(**kwargs)

    parser.set_defaults(func=run)

    return parser


#
# Inspect
#
def prepare_pod5_inspect_argparser(
    parent: Optional[argparse._SubParsersAction] = None,
) -> argparse.ArgumentParser:
    """Create an argument parser for the pod5 inspect tool"""

    _desc = "Inspect the contents of a pod5 file"
    if parent is None:
        parser = argparse.ArgumentParser(description=_desc)
    else:
        parser = parent.add_parser(
            name="inspect",
            description=_desc,
            epilog="Example: pod5 inspect reads input.pod5",
        )

    def run(**kwargs):
        from pod5.tools.pod5_inspect import inspect_pod5

        return inspect_pod5(**kwargs)

    subparser = parser.add_subparsers(title="command", dest="command")
    summary_parser = subparser.add_parser(
        "summary",
        description="Print a summary of the contents of pod5 files",
        epilog="Example: pod5 inspect summary input.pod5",
    )
    summary_parser.add_argument("input_files", type=Path, nargs="+")
    summary_parser.set_defaults(func=run)

    reads_parser = subparser.add_parser(
        "reads",
        description="Print read information on all reads as a csv table",
        epilog="Example: pod5 inspect reads input.pod5",
    )
    reads_parser.add_argument("input_files", type=Path, nargs="+")
    add_recursive_argument(reads_parser)
    reads_parser.set_defaults(func=run)

    read_parser = subparser.add_parser(
        "read",
        description="Print detailed read information for a named read id",
        epilog="Example: pod5 inspect read input.pod5 0000173c-bf67-44e7-9a9c-1ad0bc728e74",
    )
    read_parser.add_argument("input_files", type=Path, nargs=1)
    read_parser.add_argument("read_id", type=str)
    read_parser.set_defaults(func=run, recursive=False)

    debug_parser = subparser.add_parser(
        "debug",
        description="Print debugging information",
        epilog="Example: pod5 inspect debug input.pod5",
    )
    debug_parser.add_argument("input_files", type=Path, nargs=1)
    debug_parser.set_defaults(func=run)

    return parser


#
# MERGE
#
def prepare_pod5_merge_argparser(
    parent: Optional[argparse._SubParsersAction] = None,
) -> argparse.ArgumentParser:
    """Create an argument parser for the pod5 merge tool"""

    _desc = "Merge multiple pod5 files"

    if parent is None:
        parser = argparse.ArgumentParser(description=_desc)
    else:
        parser = parent.add_parser(
            name="merge",
            description=_desc,
            formatter_class=SubcommandHelpFormatter,
            epilog="Example: pod5 merge inputs/*.pod5 merged.pod5",
        )

    # Core arguments
    parser.add_argument(
        "inputs",
        type=Path,
        nargs="+",
        help="Pod5 filepaths to use as inputs",
    )
    parser.add_argument(
        "-o",
        "--output",
        required=True,
        type=Path,
        help="Output filepath",
    )
    add_recursive_argument(parser)
    add_force_overwrite_argument(parser)
    parser.add_argument(
        "-t",
        "--threads",
        type=int,
        default=DEFAULT_THREADS,
        help="Number of workers",
    )
    parser.add_argument(
        "-R",
        "--readers",
        type=int,
        default=20,
        help="number of merge readers TESTING ONLY",
    )
    parser.add_argument(
        "-D",
        "--duplicate-ok",
        action="store_true",
        help="Allow duplicate read_ids",
    )

    def run(**kwargs):
        from pod5.tools.pod5_merge import merge_pod5

        return merge_pod5(**kwargs)

    parser.set_defaults(func=run)
    return parser


#
# Repack
#
def prepare_pod5_repack_argparser(
    parent: Optional[argparse._SubParsersAction] = None,
) -> argparse.ArgumentParser:
    """Create an argument parser for the pod5 repack tool"""

    _desc = "Repack a pod5 files into a single output"
    if parent is None:
        parser = argparse.ArgumentParser(description=_desc)
    else:
        parser = parent.add_parser(
            name="repack",
            description=_desc,
            epilog="Example: pod5 repack inputs/*.pod5 repacked/",
        )

    parser.add_argument(
        "inputs", type=Path, nargs="+", help="Input pod5 file(s) to repack"
    )
    parser.add_argument(
        "-o", "--output", type=Path, help="Output directory for pod5 files"
    )
    add_recursive_argument(parser)
    add_force_overwrite_argument(parser)
    parser.add_argument(
        "-t",
        "--threads",
        type=int,
        default=DEFAULT_THREADS,
        help="Number of repacking workers",
    )

    def run(**kwargs):
        from pod5.tools.pod5_repack import repack_pod5

        return repack_pod5(**kwargs)

    parser.set_defaults(func=run)
    return parser


#
# Subset
#
def prepare_pod5_subset_argparser(
    parent: Optional[argparse._SubParsersAction] = None,
) -> argparse.ArgumentParser:
    """Create an argument parser for the pod5 subset tool"""

    _desc = (
        "Given one or more pod5 input files, take subsets of reads "
        "into one or more pod5 output files by a user-supplied mapping."
    )
    if parent is None:
        parser = argparse.ArgumentParser(description=_desc)
    else:
        parser = parent.add_parser(
            name="subset",
            description=_desc,
            formatter_class=SubcommandHelpFormatter,
            epilog="Example: pod5 subset inputs.pod5 --output subset_mux/ "
            "--summary summary.tsv --columns mux",
        )

    # Core arguments
    parser.add_argument(
        "inputs", type=Path, nargs="+", help="Pod5 filepaths to use as inputs"
    )
    parser.add_argument(
        "-o",
        "--output",
        type=Path,
        default=Path.cwd(),
        help="Destination directory to write outputs",
    )
    add_recursive_argument(parser)
    add_force_overwrite_argument(parser)
    parser.add_argument(
        "-t",
        "--threads",
        type=int,
        default=DEFAULT_THREADS,
        help="Number of subsetting workers",
    )

    mapping_group = parser.add_argument_group("direct mapping")
    mapping_exclusive = mapping_group.add_mutually_exclusive_group(required=False)
    mapping_exclusive.add_argument(
        "--csv",
        type=Path,
        help="CSV file mapping output filename to read ids",
    )

    table_group = parser.add_argument_group("table mapping")

    # Allow --summary or --table
    table_group.add_argument(
        "-s",
        "--summary",
        "--table",
        type=Path,
        help="Table filepath (csv or tsv)",
        dest="table",
    )
    table_group.add_argument(
        "-R",
        "--read-id-column",
        type=str,
        default="read_id",
        help="Name of the read_id column in the summary",
    )
    table_group.add_argument(
        "-c",
        "--columns",
        type=str,
        nargs="+",
        help="Names of --summary / --table columns to subset on",
    )
    table_group.add_argument(
        "--template",
        type=str,
        default=None,
        help="template string to generate output filenames "
        '(e.g. "mux-{mux}_barcode-{barcode}.pod5"). '
        "default is to concatenate all columns to values as shown in the example.",
    )
    table_group.add_argument(
        "-T",
        "--ignore-incomplete-template",
        action="store_true",
        default=None,
        help="Suppress the exception raised if the --template string does not contain "
        "every --columns key",
    )

    content_group = parser.add_argument_group("content settings")
    content_group.add_argument(
        "-M",
        "--missing-ok",
        action="store_true",
        help="Allow missing read_ids",
    )
    content_group.add_argument(
        "-D",
        "--duplicate-ok",
        action="store_true",
        help="Allow duplicate read_ids",
    )

    def run(**kwargs):
        from pod5.tools.pod5_subset import subset_pod5

        return subset_pod5(**kwargs)

    parser.set_defaults(func=run)

    return parser


#
# Recover
#
def prepare_pod5_recover_argparser(
    parent: Optional[argparse._SubParsersAction] = None,
) -> argparse.ArgumentParser:
    """Create an argument parser for the pod5 recover tool"""

    _desc = (
        "Attempt to recover pod5 files. Recovered files are written "
        "to sibling files with the '_recovered.pod5` suffix"
    )
    if parent is None:
        parser = argparse.ArgumentParser(description=_desc)
    else:
        parser = parent.add_parser(name="recover", description=_desc)

    parser.add_argument(
        "inputs", type=Path, nargs="+", help="Input pod5 file(s) to update"
    )

    add_recursive_argument(parser)
    add_force_overwrite_argument(parser)

    def run(**kwargs) -> Any:
        from pod5.tools.pod5_recover import recover_pod5

        return recover_pod5(**kwargs)

    parser.set_defaults(func=run)

    return parser


#
# Update
#
def prepare_pod5_update_argparser(
    parent: Optional[argparse._SubParsersAction] = None,
) -> argparse.ArgumentParser:
    """Create an argument parser for the pod5 update tool"""

    _desc = "Update a pod5 files to the latest available version"
    if parent is None:
        parser = argparse.ArgumentParser(description=_desc)
    else:
        parser = parent.add_parser(
            name="update",
            description=_desc,
            formatter_class=SubcommandHelpFormatter,
        )

    parser.add_argument(
        "inputs", type=Path, nargs="+", help="Input pod5 file(s) to update"
    )
    parser.add_argument(
        "-o",
        "--output",
        type=Path,
        help="Output directory for updated pod5 files",
        required=True,
    )

    add_recursive_argument(parser)
    add_force_overwrite_argument(parser)

    def run(**kwargs) -> Any:
        from pod5.tools.pod5_update import update_pod5

        return update_pod5(**kwargs)

    parser.set_defaults(func=run)

    return parser


#
# View
#
def prepare_pod5_view_argparser(
    parent: Optional[argparse._SubParsersAction] = None,
) -> argparse.ArgumentParser:
    """Create an argument parser for the pod5 view tool"""

    _desc = """
    Write contents of some pod5 file(s) as a table to stdout or --output if given.
    The default separator is <tab>.
    The column order is always as shown in -L/--list-fields"
    """

    if parent is None:
        parser = argparse.ArgumentParser(description=_desc)
    else:
        parser = parent.add_parser(
            name="view",
            description=_desc,
            epilog="Example: pod5 view input.pod5",
            formatter_class=SubcommandHelpFormatter,
        )

    parser.add_argument(
        "inputs", type=Path, nargs="*", help="Input pod5 file(s) to view"
    )

    parser.add_argument(
        "-o", "--output", type=Path, default=None, help="Output filename"
    )
    add_recursive_argument(parser)
    add_force_overwrite_argument(parser)
    parser.add_argument(
        "-t",
        "--threads",
        default=DEFAULT_THREADS,
        type=int,
        help="Set the number of reader workers",
    )
    format_group = parser.add_argument_group("Formatting")
    format_group.add_argument(
        "-H", "--no-header", action="store_true", help="Omit the header line"
    )
    format_group.add_argument(
        "--separator",
        default="\t",
        help="Table separator character (e.g. ',')",
        type=str,
    )

    selection = parser.add_argument_group("Selection")
    selection.add_argument(
        "-I",
        "--ids",
        action="store_true",
        help="Only write 'read_id' field",
        dest="group_read_id",
    )
    selection.add_argument(
        "-i",
        "--include",
        type=str,
        help="Include a double-quoted comma-separated list of fields",
    )
    selection.add_argument(
        "-x",
        "--exclude",
        type=str,
        help="Exclude a double-quoted comma-separated list of fields.",
    )

    help_group = parser.add_argument_group("List Fields")
    help_group.add_argument(
        "-L",
        "--list-fields",
        action="store_true",
        help="List all groups and fields available for selection and exit",
    )

    def run(**kwargs):
        from pod5.tools.pod5_view import view_pod5

        return view_pod5(**kwargs)

    parser.set_defaults(func=run)

    return parser



"""
Tool for inspecting the contents of pod5 files
"""
import csv
import os
import sys
from dataclasses import asdict
from pathlib import Path
from typing import Callable, Dict, List
from uuid import UUID

import pod5 as p5
from pod5.tools.parsers import prepare_pod5_inspect_argparser, run_tool
from pod5.tools.utils import collect_inputs


def format_shift_scale_pair(pair):
    return f"({pair.shift} {pair.scale})"


def format_shift_scale_pair_num(pair):
    return f"({pair.shift:.1f} {pair.scale:.1f})"


def do_reads_command(reader: p5.Reader, write_header: bool):
    keys = [
        "read_id",
        "channel",
        "well",
        "pore_type",
        "read_number",
        "start_sample",
        "end_reason",
        "median_before",
        "num_samples",
        "byte_count",
        "signal_compression_ratio",
        "num_minknow_events",
        "tracked_scaling",
        "predicted_scaling",
        "num_reads_since_mux_change",
        "time_since_mux_change",
    ]

    csv_read_writer = csv.DictWriter(sys.stdout, keys)

    # Only write header on first call
    if write_header:
        csv_read_writer.writeheader()

    for read in reader.reads():
        fields = {
            "read_id": read.read_id,
            "channel": read.pore.channel,
            "well": read.pore.well,
            "pore_type": read.pore.pore_type,
            "read_number": read.read_number,
            "start_sample": read.start_sample,
            "end_reason": read.end_reason.name,
            "median_before": f"{read.median_before:.1f}",
            "num_samples": read.num_samples,
            "byte_count": read.byte_count,
            "signal_compression_ratio": f"{read.byte_count / float(read.sample_count*2):.3f}",
            "num_minknow_events": read.num_minknow_events,
            "tracked_scaling": format_shift_scale_pair_num(read.tracked_scaling),
            "predicted_scaling": format_shift_scale_pair_num(read.predicted_scaling),
            "num_reads_since_mux_change": read.num_reads_since_mux_change,
            "time_since_mux_change": read.time_since_mux_change,
        }

        try:
            csv_read_writer.writerow(fields)
        except BrokenPipeError:
            devnull = os.open(os.devnull, os.O_WRONLY)
            os.dup2(devnull, sys.stdout.fileno())
            break


def dump_run_info(run_info: p5.RunInfo):
    tab = "\t"
    for name, value in asdict(run_info).items():
        if isinstance(value, list):
            print(f"{tab}{name}:")
            for k, v in value:
                print(f"{tab*2}{k}: {v}")
        else:
            print(f"{tab}{name}: {value}")


def do_read_command(reader: p5.Reader, read_id: str, **_):
    try:
        uuid_read_id = UUID(read_id)

    except ValueError:
        print(f"Supplied read_id '{read_id}' is not a valid UUID", file=sys.stderr)
        return

    for read in reader.reads():
        if read.read_id != uuid_read_id:
            continue

        print(f"read_id: {read.read_id}")
        print(f"read_number:\t{read.read_number}")
        print(f"start_sample:\t{read.start_sample}")
        print(f"median_before:\t{read.median_before}")
        print("channel data:")
        print(f"\tchannel: {read.pore.channel}")
        print(f"\twell: {read.pore.well}")
        print(f"\tpore_type: {read.pore.pore_type}")
        print("end reason:")
        print(f"\tname: {read.end_reason.name}")
        print(f"\tforced: {read.end_reason.forced}")
        print("calibration:")
        print(f"\toffset: {read.calibration.offset}")
        print(f"\tscale: {read.calibration.scale}")
        print("samples:")
        print(f"\tsample_count: {read.sample_count}")
        print(f"\tbyte_count: {read.byte_count}")
        print(
            f"\tcompression ratio: {read.byte_count / float(read.sample_count*2):.3f}"
        )

        print("run info:")
        dump_run_info(read.run_info)
        break


def do_debug_command(reader: p5.Reader, **_):
    batch_count = 0
    batch_sizes = []
    read_count = 0
    sample_count = 0
    byte_count = 0
    min_sample = float("inf")
    max_sample = 0

    run_infos = {}

    for batch in reader.read_batches():
        batch_count += 1

        batch_read_count = 0
        for read in batch.reads():
            batch_read_count += 1
            read_sample_count = read.sample_count
            sample_count += read_sample_count
            byte_count += read.byte_count

            run_info_index = read.run_info_index
            if run_info_index not in run_infos:
                run_infos[run_info_index] = read.run_info

            min_sample = min(min_sample, read.start_sample)
            max_sample = max(max_sample, read.start_sample + read_sample_count)
        batch_sizes.append(batch_read_count)
        read_count += batch_read_count

    print(f"Contains {read_count} reads, in {batch_count} batches: {batch_sizes}")
    print(f"Reads span from sample {min_sample} to {max_sample}")
    print(
        f"{sample_count} samples, {byte_count}"
        f" bytes: {100*byte_count/float(sample_count*2):.1f} % signal compression ratio"
    )

    for idx, run_info in run_infos.items():
        print(f"Run info {idx}:")
        dump_run_info(run_info)


def do_summary_command(reader: p5.Reader, **kwargs):
    batch_count = 0
    total_read_count = 0

    print(
        f"File version in memory {reader.file_version}, read table version {reader.reads_table_version}."
    )
    print(f"File version on disk {reader.file_version_pre_migration}.")
    if reader.is_vbz_compressed:
        print("File uses VBZ compression.")
    else:
        print("File is uncompressed.")

    for batch in reader.read_batches():
        batch_count += 1

        batch_read_count = 0
        for _ in batch.reads():
            batch_read_count += 1

        print(f"Batch {batch_count}, {batch_read_count} reads")
        total_read_count += batch_read_count
    print(f"Found {batch_count} batches, {total_read_count} reads")


def inspect_pod5(
    command: str, input_files: List[Path], recursive: bool = False, **kwargs
):
    """Determine which inspect command to run from the parsed arguments and run it"""

    commands: Dict[str, Callable] = {
        "reads": do_reads_command,
        "read": do_read_command,
        "summary": do_summary_command,
        "debug": do_debug_command,
    }

    for idx, filename in enumerate(
        collect_inputs(input_files, recursive=recursive, pattern="*.pod5")
    ):
        try:
            reader = p5.Reader(filename)
        except Exception as exc:
            print(f"Failed to open pod5 file: {filename}: {exc}", file=sys.stderr)
            continue

        kwargs["reader"] = reader
        kwargs["write_header"] = idx == 0
        commands[command](**kwargs)


def main():
    """Run the pod5 inspect tool"""
    run_tool(prepare_pod5_inspect_argparser())


if __name__ == "__main__":
    main()
